from pylab import array, isvector, vstack, hstack, zeros, ones, sqrt, sum, exp, log
from pylab import linspace, arange, pi, meshgrid, cos, sin, arccos, angle, tan, atleast_2d
from pylab import absolute, arctan, histogram2d, vectorize, kron, real, imag, pad, amax

from scipy import ndimage
import plotly.graph_objs as go
from plotly import figure_factory as FF
from IPython.display import HTML, display
from plotly.offline import iplot
from plotly.figure_factory import create_trisurf as trisurf

import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.cm as cm

from skimage.measure import find_contours

class KendallTriangles :
	def __init__(self, mode = 'whole sphere', resolution = (45, 90) ):
		"mode = 'whole sphere' | 'spherical blackboard'"
		self.mode = mode
		if self.mode == 'whole sphere' :
			thetas = linspace(0,   pi, resolution[0] + 1)
			phis   = linspace(-pi, pi, resolution[1] + 1)
		elif self.mode == 'spherical blackboard' :
			thetas = linspace(0,.5*pi, resolution[0] + 1)
			phis   = linspace(-pi/3, 0, resolution[1] + 1)
			
		(self.theta_grid, self.phi_grid) = meshgrid(thetas, phis)
		self.theta_centers = (self.theta_grid[0:-1,0:-1] + self.theta_grid[0:-1,1:] + self.theta_grid[1:,0:-1] + self.theta_grid[1:,1:]) / 4.
		self.phi_centers   = (self.phi_grid[0:-1,0:-1]   + self.phi_grid[0:-1,1:]   + self.phi_grid[1:,0:-1]   + self.phi_grid[1:,1:] ) / 4.
		
		areas_thetas = (phis[1]-phis[0]) * ( -cos(thetas[1:]) + cos(thetas[0:-1]))
		self.cell_areas = vstack( (areas_thetas,)*(resolution[1]) ).T
		
		# Setup routines for the interactive "Plot.ly" display
		self.setup_layout()
		self.current_axis = []
	
	def generate_spherical_sampled_data(self) :
		Theta = []
		Phi   = []
		
		for theta in linspace(0,pi,17*2) :
			for phi in arange(0, 2*pi, (2*pi)/48) :
				Theta.append(theta)
				Phi.append(phi)
		Theta = array(Theta)
		Phi   = array(Phi)
		
		if True :
			# Recompute X, Y, Z
			X = sin(Theta) * cos(Phi)
			Y = sin(Theta) * sin(Phi)
			Z = cos(Theta)
			# Recompute Phi, Theta
			Theta = arccos(X)
			Phi   = angle(Y + 1j*Z)
		
		M = sqrt(3) * tan(Theta/2)
		triangles = vstack( (
			-1 * ones( (len(Theta),) ),
			+1 * ones( (len(Theta),) ),
			M*exp(1j*Phi)
		) ).T
		
		self.set_data(triangles)
	
	def show_pieces(self) :
		def aux(P, color, name) :
			curves = go.Scatter3d(x = P[:,0], y = P[:,1], z = P[:,2], mode = 'lines', hoverinfo='none',
								  line   = dict(width= 3, color=color), name = name)
			self.current_axis.append(curves)
		
		R = .501
		T = linspace(0,2*pi,100)
		
		Eq = vstack(( cos(T), sin(T), 0*T )).T
		aux(R*Eq, 'red', 'Flat triangles')
		
		M1 = vstack(( cos(T), 0*T, sin(T) )).T
		aux(R*M1, 'blue', 'AC = BC')
		
		M2 = vstack(( cos(T)/2, cos(T)*sqrt(3)/2, sin(T) )).T
		aux(R*M2, 'green', 'AB = AC')
		
		M3 = vstack(( -cos(T)/2, cos(T)*sqrt(3)/2, sin(T) )).T
		aux(R*M3, 'cyan', 'AB = CB')
		
	def show_xi(self,scale = .03):
		"Displays the dataset in the xi coordinate."
		
		
		def aux(P, color, name) :
			curves = go.Scatter(x = P[:,0], y = P[:,1], mode = 'lines', hoverinfo='none',
								  line   = dict(width= 3, color=color), name = name)
			self.current_axis.append(curves)
		
		T = linspace(-pi,pi,100)
		
		XYZ_2 = vstack(( cos(T)/2, cos(T)*sqrt(3)/2, sin(T) )).T
		Theta_2 = arccos(XYZ_2[:,0])
		Phi_2   = angle(XYZ_2[:,1] + 1j*XYZ_2[:,2])
		R_2     = tan(Theta_2/2)
		M2 = vstack(( R_2*cos(Phi_2), R_2*sin(Phi_2) )).T
		aux(M2, 'green', 'AB = AC')
		
		XYZ_3 = vstack(( -cos(T)/2, cos(T)*sqrt(3)/2, sin(T) )).T
		Theta_3 = arccos(XYZ_3[:,0])
		Phi_3   = angle(XYZ_3[:,1] + 1j*XYZ_3[:,2])
		R_3     = tan(Theta_3/2)
		M3 = vstack(( R_3*cos(Phi_3), R_3*sin(Phi_3) )).T
		aux(M3, 'cyan', 'AB = CB')
		
		M1 = vstack(( 0*T, T )).T
		aux(M1, 'blue', 'AC = BC')
		
		Eq = vstack(( T, 0*T)).T
		aux(Eq, 'red', 'Flat triangles')
		
		
		
		
		# We don't want glyphs to overlap
		e_theta = scale * array([0.,1.])
		e_phi   = scale * array([1.,0.])
		
		glyphs = []
		separator = [None, None]
		for i in range(self.triangles.shape[0]) :
			point = array([self.R[i]*cos(self.phi[i]), self.R[i]*sin(self.phi[i])])
			glyphs.append(array([
				point + real(self.triangles[i,0]) * e_phi + imag(self.triangles[i,0]) * e_theta , # A
				point + real(self.triangles[i,1]) * e_phi + imag(self.triangles[i,1]) * e_theta , # B
				point + real(self.triangles[i,2]) * e_phi + imag(self.triangles[i,2]) * e_theta , # C
				point + real(self.triangles[i,0]) * e_phi + imag(self.triangles[i,0]) * e_theta , # A
				separator
			]))
		glyphs = vstack(glyphs)
		curves = go.Scatter(x = glyphs[:,0], y = glyphs[:,1], mode = 'lines', hoverinfo='none', name = 'Triangles', line = dict( color = 'black', width=1))
		self.current_axis.append(curves)
		
		
		
		axis = dict(
				showbackground=True,
				backgroundcolor='rgb(230, 230, 230)',
				gridcolor='rgb(255, 255, 255)',
				zerolinecolor='rgb(255, 255, 255)'
			)
		#aspectratio=dict(x=1, y=1)
		layout = go.Layout(
			title='Triangles',
			width=1000,
			height= 1000,
			xaxis=dict(
				range=[-2, 2]
			),
			yaxis=dict(
				range=[-2, 2]
			)
			)
		(wid, uid) =  iplot(go.Figure(data=self.current_axis, layout=layout))
		display(wid)
		
		
	def plot_data(self, N_triangles = 10, sigma = None) :
		"Displays first triangles using matplotlib."
		A = 10
		B = 5
		plt.figure(figsize=(10, 6), dpi=100, facecolor='w', edgecolor='k')      
		
		if sigma == "unit disc" :
			x,y  = meshgrid(linspace(-A,A,100), linspace(-B,B,100))
			z    = .75*(x**2 + y**2 <= 1)                               
			plt.imshow(z, interpolation='bilinear', origin='lower', 
					vmin = 0, vmax = 1, cmap=cm.gist_yarg, 
					extent=(-A, A, -B, B)) 
		elif sigma is not None :
			x,y  = meshgrid(linspace(-A,A,100), linspace(-B,B,100))
			z    = .75*exp(- .5*((x/sigma)**2 + y**2))                               
			plt.imshow(z, interpolation='bilinear', origin='lower', 
					vmin = 0, vmax = 1, cmap=cm.gist_yarg, 
					extent=(-A, A, -B, B)) 
			                            
		#plt.colorbar() 
		
		cNorm  = colors.Normalize(vmin=0, vmax=N_triangles-1)
		scalarMap = cm.ScalarMappable(norm=cNorm, cmap=plt.get_cmap('rainbow') )
		
		plt.autoscale(tight=True)
		
		for (i, line) in enumerate(self.raw_data[:N_triangles,:]) :
			line = array([line[0],line[1],line[2],line[0]])
			x = real(line)
			y = imag(line)
			plt.plot( x, y , color=scalarMap.to_rgba(i))
		
		plt.axhline(0, color='black')
		plt.axvline(0, color='black')
		plt.axis([-A, A, -B, B])
		plt.axis('equal')
		plt.show()
		
		
		
	def set_data(self, triangles):
		triangles = atleast_2d(triangles)
		assert (triangles.shape[1] == 3), 'Triangles should be given as triplets of complex numbers.'
		self.raw_data = triangles
		A = triangles[:,0]
		B = triangles[:,1]
		C = triangles[:,2]
		
		# See section 5 : The manifold carrying the shapes of triangles
		unit_length = absolute(B-A)/2
		self.M = absolute(C - (A+B)/2) / unit_length
		self.psi = angle( (C - .5*(A+B)) / (B-A) )
		# eq. (34) : 
		self.R = self.M / sqrt(3)
		self.phi = self.psi
		self.theta = 2 * arctan(self.R)
		# eq. (35)
		self.X = cos(self.theta)
		self.Y = sin(self.theta) * cos(self.phi)
		self.Z = sin(self.theta) * sin(self.phi)
		# Alternative spherical coordinates around the Z axis :
		self.theta_Z = arccos(self.Z)
		self.phi_Z   = angle(self.X + 1j * self.Y)
		
		if self.mode == 'spherical blackboard' :
			# Quotient by the symmetry + relabelling group
			self.theta_Z = arccos(abs(self.Z))
			self.phi_Z   = - absolute(  mod(self.phi_Z + (pi/3), 2*pi/3)  - (pi/3)  )
			# Recompute X, Y, Z
			self.X = sin(self.theta_Z) * cos(self.phi_Z)
			self.Y = sin(self.theta_Z) * sin(self.phi_Z)
			self.Z = cos(self.theta_Z)
			# Recompute Phi, Theta
			self.theta = arccos(self.X)
			self.phi   = angle(self.Y + 1j*self.Z)
			self.R     = tan(self.theta / 2)
		# eq. (36)
		A = - sqrt(3) - self.R * exp(1j * self.phi)
		B = + sqrt(3) - self.R * exp(1j * self.phi)
		C = 2 *         self.R * exp(1j * self.phi)
		self.triangles = vstack( (A,B,C) )
		
		# normalize self.triangles (for display purposes)
		centroids = sum(self.triangles, 0)
		self.triangles -= centroids
		norms = sqrt( sum( absolute(self.triangles)**2, 0) ) # inefficient, but whatever
		self.triangles /= norms
		self.triangles = self.triangles.T
		
		# Create an histogram
		self.compute_density()
		
	def compute_density(self) :
		"Computes an empirical histogram on the (theta x phi) grid."
		(self.density, xedges, yedges) = histogram2d(self.theta_Z, self.phi_Z, bins = (self.theta_grid[0], self.phi_grid[:,0]) )
		# some smoothing to prevent over-sensitivity next to the poles
		if self.mode == 'whole sphere' :
			sigma = .5
			total_area = 4*pi
			mode = 'wrap'
		elif self.mode == 'spherical blackboard' :
			sigma = 1.5
			total_area = 4*pi / 12
			mode = 'nearest'
		for (i, th) in enumerate( self.theta_centers[0] ) :
			self.density[i] = ndimage.filters.gaussian_filter1d(self.density[i], sigma / sin(th), mode = mode)
		self.density = self.density / sum(self.density)
		self.density = total_area*self.density / self.cell_areas
		
	def mesh(self, f) :
		"Returns a mesh (points, triangles, signals_per_triangle) with signals given by f."
		ntheta = self.theta_grid.shape[1]
		nphi   = self.theta_grid.shape[0]

		Z = .5 * cos(self.theta_grid)
		X = .5 * sin(self.theta_grid) * cos(self.phi_grid)
		Y = .5 * sin(self.theta_grid) * sin(self.phi_grid)
		
		
		# store the points in a npoints-by-3 array,
		# where npoints = ntheta * nphi
		points = vstack( ( X.ravel() ,
						   Y.ravel() ,
						   Z.ravel() ) ).T
		# generate the connectivity matrix, a ntriangles-by-3 int matrix,
		# each line giving the indices of 3 points -> one triangle.
		Ilen = nphi 
		Jlen = ntheta 
		#triangles = vstack([ [i+j*(ylen+1),  (i+1)+j*(ylen+1),  i+(j+1)*(ylen+1) ] for i in range(ylen) for j in range(xlen)])
		triangles = vstack( [  [ ( j +     Jlen*i    , (j+1) + Jlen*i     , j +     Jlen*(i+1) ) for j in range(Jlen - 1) ]
							 + [ ( j + 1 + Jlen*(i+1), j     + Jlen*(i+1) , j + 1 + Jlen*i     ) for j in range(Jlen - 1) ]
							   for i in range(Ilen - 1) ] )
		if f == 'density' :
			signals = self.density
		else :
			fv = vectorize(f)
			signals = fv(self.theta_centers.T, self.phi_centers.T)
		# given the connectivity matrix 'triangles', 
		# and the value of f on rectangles,
		# signals_per_triangle is easy to compute
		signals = kron( signals.T, [[1], [1]] )
		signals_per_triangle = signals.ravel()
		
		return (points, triangles, signals_per_triangle)
		
	def clear_axis(self):
		self.current_axis = []
	def show(self) : 
		"Interactive Display of the Kendall manifold."
		self.show_function(lambda theta,phi : 1., faces = True, contours = False, colormap = "YlOrRd")
		
	def show_glyphs(self, scale = 0.03, axis = 'Z') :
		"Displays triangles on the spherical manifold."
		if self.mode == 'whole sphere' or self.mode == 'spherical blackboard' :
			# We will embed self.triangles in the euclidean space R^3,
			# in the neighborhood of the sphere S(1/2).
			
			if axis == 'X' :
				theta = self.theta
				phi   = self.phi
				e_theta = vstack( ( -sin(theta)        ,  cos(theta) * cos(phi), cos(theta) * sin(phi) ) ).T
				e_phi   = vstack( ( zeros(theta.shape) , -sin(theta) * sin(phi), sin(theta) * cos(phi) ) ).T
			elif axis == 'Z' :
				theta = self.theta_Z
				phi   = self.phi_Z
				e_theta = - vstack( (  cos(theta) * cos(phi), cos(theta) * sin(phi), -sin(theta)         ) ).T
				e_phi   = + vstack( ( - sin(phi),  cos(phi), zeros(theta.shape)  ) ).T
				
			# We don't want glyphs to overlap
			e_theta = scale * e_theta
			e_phi   = scale * e_phi
			
			glyphs = []
			separator = [None, None, None]
			for i in range(self.triangles.shape[0]) :
				point = .502 * array([self.X[i], self.Y[i], self.Z[i]])
				T = self.triangles[i,:]
				glyphs.append(array([
					point + real(T[0]) * e_phi[i] + imag(T[0]) * e_theta[i] , # A
					point + real(T[1]) * e_phi[i] + imag(T[1]) * e_theta[i] , # B
					point + real(T[2]) * e_phi[i] + imag(T[2]) * e_theta[i] , # C
					point + real(.6*T[0] + .4*T[2]) * e_phi[i] + imag(.6*T[0] + .4*T[2]) * e_theta[i] , # .8*A + .2*C
					separator
				]))
			glyphs = vstack(glyphs)
			curves = go.Scatter3d(x = glyphs[:,0], y = glyphs[:,1], z = glyphs[:,2], mode = 'lines', hoverinfo='none', name = 'Triangles', line = dict( color = 'black'))
			self.current_axis.append(curves)
			
	def show_markers(self, **kwargs) :
		"Marker plot of the data on the spherical manifold."
		None


	def add_markers(self, triangles, color="black", name = "", scale=1.) :
		"Marker plot of the data on the spherical manifold."

		if triangles.dim() == 2 :
			triangles = triangles.detach().numpy()
			triangles = triangles[:,0] + 1j*triangles[:,1]
			triangles = atleast_2d(triangles)
		elif triangles.dim() == 3 :
			triangles = triangles.detach().numpy()
			triangles = triangles[:,:,0] + 1j*triangles[:,:,1]

		assert (triangles.shape[1] == 3), 'Triangles should be given as triplets of complex numbers.'
		raw_data = triangles
		A = triangles[:,0]
		B = triangles[:,1]
		C = triangles[:,2]
		
		# See section 5 : The manifold carrying the shapes of triangles
		unit_length = absolute(B-A)/2
		M = absolute(C - (A+B)/2) / unit_length
		psi = angle( (C - .5*(A+B)) / (B-A) )
		# eq. (34) : 
		R = M / sqrt(3)
		phi = psi
		theta = 2 * arctan(R)
		# eq. (35)
		X = .502*cos(theta)
		Y = .502*sin(theta) * cos(phi)
		Z = .502*sin(theta) * sin(phi)

		dots = go.Scatter3d(x = X, y = Y, z = Z, name = name, mode="markers",
							marker= dict( color=color, size=scale*12, symbol='circle') )
		self.current_axis.append(dots)
			
	
	def show_density(self, contours = True) :
		self.show_function('density', faces = True, contours = contours)
		
	def show_function(self, f, faces = True, contours = True, name = 'f', colormap = "Hot") :
		"Displays a function f(phi, theta)."
		
		# Trisurf plot
		if faces :
			(points, triangles, signals_per_triangle) = self.mesh(f)
			points = array(points)
			triangles = array(triangles)
			#signals = array(signals)
			signals_per_triangle = array(signals_per_triangle)
			#signals_per_triangle = list( (signals[triangles[i,0]] + signals[triangles[i,1]] + signals[triangles[i,2]]) / 3
			#							for i in range(triangles.shape[0]) )
			signals_per_triangle[0] -= 0.01
			# Validate colormap
			#my_colormap = FF.validate_colors(colormap, 'tuple')
			newdata = trisurf(x=points[:,0], y=points[:,1], z=points[:,2],
								 colormap   = colormap,
								 simplices  = triangles,
								 color_func = signals_per_triangle,
								 plot_edges = False,
								 edges_color = 'rgb(50, 50, 50)',
								 show_colorbar = True,
								 #data_list = True,
								 #minmax_values = (0, max(signals_per_triangle) )
								 )
			self.current_axis.append(newdata.data[0])
		
		# 3D contour plot
		if contours :
			R = .501
			if f == 'density' :
				name = 'Density'
				if self.mode == 'whole sphere' :
					den = pad( pad( self.density, (0,1), 'wrap') , (1,0), 'edge')
				else :
					den = pad( self.density, 1, 'edge' )
				values = (den[1:,1:] + den[0:-1,1:] + den[1:,0:-1] + den[0:-1,0:-1]) / 4.
			else :
				fv = vectorize(f)
				values = fv(self.theta_grid.T, self.phi_grid.T)
			levels = linspace(0, amax(values), 10)
			
			for level in levels :
				contours = find_contours(values, level)
				ntheta = self.theta_grid.shape[1] - 1
				nphi   = self.theta_grid.shape[0] - 1
				if self.mode == 'whole sphere' :
					theta   = lambda x : pi* x[0] / ntheta
					phi = lambda x : pi* (2*x[1]/nphi-1)
				elif self.mode == 'spherical blackboard' :
					theta   = lambda x : (pi/2) * x[0]/ntheta
					phi     = lambda x : (pi/3) * (x[1] / nphi - 1)
				#points3D = [ ( [ ( R* cos(pi* thph[0]/ntheta), 
				#				   R* sin(pi* thph[0]/ntheta) * cos(pi*(2*thph[1]/nphi-1)), 
				#				   R* sin(pi* thph[0]/ntheta) * sin(pi*(2*thph[1]/nphi-1)) ) 
				#				 for thph in contour ] 
				#			 + [(None, None, None)] ) 
				#			for contour in contours]
				points3D = [ ( [ (  
								   R* sin(theta(thph)) * cos(phi(thph)), 
								   R* sin(theta(thph)) * sin(phi(thph)) ,
								   R* cos(theta(thph)) ) 
								 for thph in contour ] 
							 + [(None, None, None)] ) 
							for contour in contours]
				if points3D != [] :
					contours3D = vstack(points3D)
					curves = go.Scatter3d(x = contours3D[:,0], y = contours3D[:,1], z = contours3D[:,2], mode = 'lines', hoverinfo='none',
										  line   = dict(width= 3, color='red'), name = (name + ' = ' + "{:.2f}".format(level)))
					self.current_axis.append(curves)
				
	def setup_layout(self) :
		"Setup the axis properties."
		axis = {
				"showbackground"  : True,
				"backgroundcolor" : 'rgb(230, 230, 230)',
				"gridcolor"       : 'rgb(255, 255, 255)',
				"zerolinecolor"   : 'rgb(255, 255, 255)',
				"range"           : [-.55,.55],
		}
		aspectratio  = { "x" : 1, "y" : 1, "z" : 1 }
		camera = {
			"up"     : { "x" : 0,    "y" : 0,    "z" : 1  },
			"center" : { "x" : 0,    "y" : 0,    "z" : 0  },
			"eye"    : { "x" : 1.55, "y" : 1.55, "z" : .25},
		}
		
		self.layout = go.Layout(
			title='Kendall Triangles',
			width = 550,
			height= 550,
			scene = {
				"xaxis" : axis.copy(),
				"yaxis" : axis.copy(),
				"zaxis" : axis.copy(),
				"aspectratio" : aspectratio,
				"camera"      : camera,
			}
			)
			
	def iplot(self, title = '') :
		"Interactive manifold display "
		self.layout['title'] = title
		iplot(go.Figure(data=self.current_axis, layout=self.layout))

